# -*- coding: utf-8 -*- 
# @Time    : 06/03/2024 
# @Author  : nzooherd 
# @File    : resource_service.py 
# @Software: Visual Studio Code

import json
import os
import random
import xml.etree.ElementTree as ET
import opencc
from nzooherd_pyzen.classes import Singleton
from logging import getLogger

from aspect.task import task_logger
from component.object_store import ObjectStore
from component.pornherd_bot import mybot
from component.video_factory import VideoFactory
from config import Config
from dao.model import Video, VideoTags
from dao.video_dao import VideoDao
from dao.video_tags_dao import VideoTagsDao
from spider.javbus import fetch_movie_info


logger = getLogger()

class ResourceService(metaclass=Singleton):

    S3_RAW_KEY_PREFIX = "videos/raw"
    S3_PREVIEW_KEY_PREFIX = "videos/preview"

    RAW_PATH = os.path.join(Config.RESOURCE_ROOT_PATH, "raw")
    PREVIEW_PATH = os.path.join(Config.RESOURCE_ROOT_PATH, "preview")
    UPLOAD_PATH = os.path.join(Config.RESOURCE_ROOT_PATH, "upload")
    METADATA_PATH = os.path.join(Config.RESOURCE_ROOT_PATH, "metadata")
    DONE_PATH = os.path.join(Config.RESOURCE_ROOT_PATH, "done")


    def __init__(self) -> None:
        self.video_factory = VideoFactory()
        self.video_dao = VideoDao()
        self.video_tags_dao = VideoTagsDao()
        self.object_store = ObjectStore()
        self.pikpak_cache_dirs = set()
        self.converter = opencc.OpenCC('t2s')

        with open(Config.RESOURCE_PIKPAK_DIR_PATH, "r") as f:
            for line in f.readlines():
                self.pikpak_cache_dirs.add(line.strip())

    async def new_video(self, file_path: str):
        """
        New video
        """
        file_name = os.path.basename(file_path)
        video_md5 = self.video_factory.md5sum(file_path)

        for path in [self.PREVIEW_PATH, self.UPLOAD_PATH, self.METADATA_PATH]:
            hard_link_path = os.path.join(path, file_name)
            if os.path.exists(hard_link_path):
                os.remove(hard_link_path)
            os.link(file_path, hard_link_path)

        try:
            await self.video_dao.insert(Video(bangou=file_name, md5=video_md5, filename=file_name,
                                          object_key = f"{self.S3_RAW_KEY_PREFIX}/{video_md5}", 
                                          star=5))
        except Exception as e:
            logger.error(f"Insert video failed: {e}")

        os.remove(file_path)
   
    async def refresh_video(self, file_path: str):
        logger.info(f"Refresh video: {file_path}")
        video_md5 = self.video_factory.md5sum(file_path)
        video_model = await self.video_dao.get_video_by_md5(video_md5)
        if not video_model:
            return await self.new_video(file_path)

        if not video_model.online_url:
            await self._update_online_url(video_md5, file_path)
        if not video_model.preview_url:
            await self._update_preview_url(video_md5, file_path)
        if not video_model.magnet_url:
            await self._update_magnet_url(video_md5, file_path)

        if not video_model.name:
            pass

    @task_logger
    async def update_online_url(self, file_path: str):
        """
        """
        logger.info(f"Update online url: {file_path}")
        filename = os.path.basename(file_path)
        video = await self.get_only_video_by_filename(filename)
        if video.online_url:
            os.remove(file_path)
            return
        video_md5 = video.md5
        bangou = video.bangou
        object_key = f"{self.S3_RAW_KEY_PREFIX}/{video_md5}"
        if not (await self.object_store.head_object(object_key, use_s3=False)):
            await self.object_store.upload_object(object_key, file_path, Metadata={"bangou": bangou},ContentDisposition="inline", ContentType="video/mp4",
                                                                               use_s3=False) 
        online_url = self.object_store.permant_url(object_key)
        await self.video_dao.update_attribute(video_md5, online_url=online_url, bangou=bangou, object_key=object_key)

        os.remove(file_path)

    @task_logger
    async def update_preview_url(self, file_path: str):
        """
        """
        filename: str = os.path.basename(file_path)
        video = await self.get_only_video_by_filename(filename)
        if video.preview_url:
            logger.info(f"old preview url: {video.preview_url} {filename}")
            return

        video_md5 = video.md5
        format_idx = filename.rindex(".")
        name, format = filename[:format_idx], filename[format_idx+1:]

        preview_file_path = os.path.join(self.PREVIEW_PATH, f"{name}_preview.{format}")
        if not os.path.exists(preview_file_path):
            try:
                self.video_factory.preview_video(file_path, preview_file_path)
            except Exception:
                logger.event(f"Preview video failed: {file_path}")
                return

        object_key = f"{self.S3_PREVIEW_KEY_PREFIX}/{video_md5}"
        await self.object_store.upload_object(object_key, preview_file_path, ContentDisposition="inline", ContentType="video/mp4", use_s3=True, ACL="public-read")

        preview_url = self.object_store.permant_url(object_key, use_s3=True)
        await self.video_dao.update_attribute(video_md5, preview_url=preview_url)

        await self.inline_preview_url(preview_url)
        os.remove(preview_file_path)
        os.remove(file_path)
    
    @task_logger
    async def update_metadata(self, filepath: str) -> dict:
        """
        Update video metadata
        """
        filename = os.path.basename(filepath)
        video = await self.get_only_video_by_filename(filename)
        metadata = self.extract_metadata(filepath) or fetch_movie_info(video.filename)
        logger.info(f"Metadata: {metadata}")
        await self.video_dao.update_attribute(video.md5, name="".join(random.sample(metadata["title"], 6)),
                                              bangou=metadata["bangou"])
        tags = set(map(lambda x:self.converter.convert(x), metadata["tags"]))
        with open(os.path.join(Config.RESOURCE_DIR, "video", "tags.txt"), "a", encoding='utf-8') as tagsm:
            for tag in tags:
                tagsm.write(tag + "\n")
        with open(os.path.join(Config.RESOURCE_DIR, "video", "tags.json"), "r") as tags_fp:
            tag2id = json.load(tags_fp)

        inter_tags = set(tags) & set(tag2id.keys())
        for i, tag in enumerate(inter_tags):
            if i > 4:
                break
            await self.video_tags_dao.insert_not_exist(VideoTags(video_id=video.id, tag=tag))
        tags_count = len(inter_tags)
        while tags_count < 3:
            await self.video_tags_dao.insert_not_exist(VideoTags(video_id=video.id, tag=random.choice(list(tag2id.keys()))))
            tags_count += 1
        
        os.remove(filepath)

    @task_logger
    def extract_metadata(self, filepath: str):
        """
        Extract video metadata
        """
        filename = os.path.basename(filepath)

        root_pikpak = "/home/nzooherd/pikpak/oneesan"
        folders = list(map(lambda x: os.path.join(root_pikpak, x), os.listdir("/home/nzooherd/pikpak/oneesan")))
        origin_filepath = None

        for item in self.pikpak_cache_dirs:
            if os.path.basename(item) == filename:
                origin_filepath = item
                break
        else:
            with open(Config.RESOURCE_PIKPAK_DIR_PATH, "w") as f:
                for item in folders:
                    if os.path.isdir(item):
                        folders.extend(list(map(lambda x: os.path.join(item, x), os.listdir(item))))
                        continue
                    if item not in self.pikpak_cache_dirs:
                        f.write(item + "\n")
                    self.pikpak_cache_dirs.add(item)
                    if os.path.basename(item) == filename:
                        origin_filepath = item
        
        if not origin_filepath:
            logger.event(f"Origin file not found: {filename}")
            return

        origin_filedir = os.path.dirname(origin_filepath)
        metadata_nfo_file = os.path.join(origin_filedir, "movie.nfo")
        if not os.path.exists(metadata_nfo_file):
            logger.event(f"Metadata not found: {filename}")
            return
        
        tree = ET.parse(metadata_nfo_file)
        bangou = tree.getroot().find(".//uniqueid[@defult='true']").text
        title = tree.getroot().find("title").text
        tags = [tag.text for tag in tree.getroot().findall('.//tag')]
        return {"title":title, "bangou": bangou, "tags": tags}

    @task_logger
    async def inline_preview_url(self, url: str):
        """
        enable inline preview
        """
        await mybot.me_client.send_message(self.WEBPAGE_BOT, url)


    #async def update_url(self, video_md5: str, file_path: str):
    #    """
    #    """
    #    await self.update_online_url(video_md5, file_path)
    #    await self.update_preview_url(video_md5, file_path)

    #    # move to uploaded
    #    file_name = os.path.basename(file_path)
    #    uploaded_file_path = os.path.join(Config.UPLOADED_PATH, file_name)
    #    with open(uploaded_file_path, "w") as f:
    #        f.write(video_md5)

    #    os.remove(file_path)

    #async def manual_update_magnet(self, video_md5: str, magnet_url: str):
    #    pass

    #async def mannual_update_star(self, video_md5: str, star: int):
    #    pass

    #async def mannual_update_tag(self, video_md5: str, tag: str):
    #    pass

    async def get_only_video_by_filename(self, filename: str) -> Video:
        videos = await self.video_dao.query_video_by_filename(filename)
        logger.debug(f"Videos: {videos}")
        if not videos:
            logger.event(f"Video not found: {filename}")
            # TODO(修改异常)
            raise Exception("Video not found")
        if len(videos) > 1:
            logger.event(f"Multiple videos found: {filename}")
            # TODO(修改异常)
            raise Exception("Multiple videos found")
        
        return videos[0]

    async def exist_video(self, filename: str) -> bool:
        return bool(await self.video_dao.query_video_by_filename(filename))